Timer = class(Turbine.UI.Window)
function Timer:Constructor(parent, layoutTable)
    Turbine.UI.Window.Constructor(self)
    if parent then
        self:SetParent(parent)
    end

	self.UpdateEvent = function(sender, event, args)
        if type(event)=="function" then
            event(args)
        else
            if type(event)=="table" then
                for i=1,#event do
                    if type(event[i])=="function" then
                        event[i](args)
                    end
                end
            end
        end
    end

	-- timer variables
	self.current_duration = 0
	self.max_duration = 0
	self.running = false
	self.update_callback = nil

	-- overall settings
	self.width = layoutTable.Width or 200
	self.height = layoutTable.Height or 20
	self.overal_opacity = layoutTable.Overall_Opacity or 1
	self.override_opacity = layoutTable.Override_Opacity
	self.increasing = layoutTable.Increasing
	self.use_warning = layoutTable.Use_Warning
	self.warning_colour = layoutTable.Warning_Colour or Turbine.UI.Color.Crimson
	self.warning_threshold = layoutTable.Warning_Threshold
	self.background_colour = layoutTable.Background_Colour

	-- bar settings
	self.bar_colour = layoutTable.Bar_Colour
	self.bar_background_colour = layoutTable.Bar_Background_Colour
	self.bar_height_percentage = math.ceil(math.clamp(layoutTable.Bar_Height_Percentage or 100, 0, 100) / 100)
	self.bar_border_size = layoutTable.Bar_Border_Size
	self.bar_border_colour = layoutTable.Bar_Border_Colour
	self.bar_left_to_right = layoutTable.Timer_Left_To_Right

	-- message settings
	self.text = layoutTable.Text
	self.originalText = layoutTable.Original_Text
	self.font = layoutTable.Font or Turbine.UI.Lotro.Font.FixedSys15
	self.font_colour = layoutTable.Font_Colour or Turbine.UI.Color.White
	self.font_outline = layoutTable.Font_Outline or Turbine.UI.FontStyle.Outline
	self.font_outline_colour = layoutTable.Font_Outline_Colour or Turbine.UI.Color.Black
	self.text_multineline = layoutTable.MutliLine or false
	self.text_opacity = self.override_opacity and self.overal_opacity or layoutTable.Text_Opacity or 1
	self.text_alignment = layoutTable.Text_Alignment or Turbine.UI.ContentAlignment.MiddleLeft

	-- duration settings
	self.show_duration = layoutTable.Show_Duration
	self.duration_font = layoutTable.Duration_Font or Turbine.UI.Lotro.Font.FixedSys15
	self.duration_font_colour = layoutTable.Duration_Font_Colour or Turbine.UI.Color.White
	self.duration_font_outline = layoutTable.Duration_Font_Outline or Turbine.UI.FontStyle.Outline
	self.duration_font_outline_colour = layoutTable.Duration_Font_Outline_Colour or Turbine.UI.Color.Black
	self.duration_opacity = self.override_opacity and self.overal_opacity or layoutTable.Duration_Opacity
	self.duration_alignment = layoutTable.Duration_Alignment

	-- icon settings
	self.left_icon = layoutTable.Show_Icon_Left
	self.right_icon = layoutTable.Show_Icon_Right
	self.icon = layoutTable.Icon
	self.icon_opacity = self.override_opacity and self.overal_opacity or layoutTable.Icon_Opacity

	-- border settings
	self.border_size = layoutTable.Border_Size or 0
	self.border_colour = layoutTable.Border_Colour or Turbine.UI.Color.Black
	self.border_opacity = self.override_opacity and self.overal_opacity or layoutTable.Border_Opacity or 1

	-- UI code
    self:SetVisible(true)
    self:SetMouseVisible(false)
    self:SetWantsUpdates(false)    

	-- "Border"
    self.border_window = Turbine.UI.Window()
    self.border_window:SetParent(self)
    self.border_window:SetVisible(true)
    self.border_window:SetMouseVisible(false)
    self.border_window:SetBackColor(self.border_colour)
	self.border_window:SetOpacity(self.border_opacity)

	-- Inner timer
	self.inner_window = Turbine.UI.Window()
	self.inner_window:SetParent(self.border_window)
    self.inner_window:SetVisible(true)
    self.inner_window:SetMouseVisible(false)

	-- Bar 
	self.bar_border_window = Turbine.UI.Window()
	self.bar_border_window:SetParent(self.inner_window)
	self.bar_border_window:SetVisible(true)
	self.bar_border_window:SetMouseVisible(false)
	self.bar_border_window:SetBackColor(self.bar_border_colour)

	self.bar_background = Turbine.UI.Window()
	self.bar_background:SetParent(self.bar_border_window)
	self.bar_background:SetVisible(true)
	self.bar_background:SetMouseVisible(false)
	self.bar_background:SetBackColor(self.bar_background_colour)

	self.bar = Turbine.UI.Window()
	self.bar:SetParent(self.bar_background)
	self.bar:SetVisible(true)
	self.bar:SetMouseVisible(false)
	self.bar:SetBackColor(self.bar_colour)


	-- Left Icon
	if self.left_icon then
		self.left_icon_control = Turbine.UI.Window()
		self.left_icon_control:SetParent(self.inner_window)
		self.left_icon_control:SetMouseVisible(false)
		self.left_icon_control:SetVisible(true)
	end

	-- Right Icon
	if self.right_icon then
		self.right_icon_control = Turbine.UI.Window()
		self.right_icon_control:SetParent(self.inner_window)
		self.right_icon_control:SetBackground(tonumber(self.icon) or self.icon)
		self.right_icon_control:SetMouseVisible(false)
		self.right_icon_control:SetVisible(true)
	end

	-- Timer Text
	self.text_window = Turbine.UI.Window()
	self.text_window:SetParent(self.inner_window)
    self.text_window:SetVisible(true)
    self.text_window:SetMouseVisible(false)

	self.text_label = Turbine.UI.Label()
	self.text_label:SetParent(self.text_window)
    self.text_label:SetVisible(true)
    self.text_label:SetMouseVisible(false)
    self.text_label:SetTextAlignment(self.text_alignment)
    self.text_label:SetForeColor(self.font_colour)
    self.text_label:SetFont(self.font)
    self.text_label:SetOutlineColor(self.font_outline_colour)
    self.text_label:SetFontStyle(self.font_outline)
    self.text_label:SetMultiline(self.text_multineline)
    self.text_label:SetMarkupEnabled(true)
    self.text_label:SetText(self.text)

	-- Duration
	self.duration_window = Turbine.UI.Window()
	self.duration_window:SetParent(self.inner_window)
    self.duration_window:SetVisible(true)
    self.duration_window:SetMouseVisible(false)

	self.duration_label = Turbine.UI.Label()
	self.duration_label:SetParent(self.text_window)
    self.duration_label:SetVisible(self.show_duration)
    self.duration_label:SetMouseVisible(false)
    self.duration_label:SetTextAlignment(self.duration_alignment)
    self.duration_label:SetForeColor(self.duration_font_colour)
    self.duration_label:SetFont(self.duration_font)
    self.duration_label:SetOutlineColor(self.duration_font_outline_colour)
    self.duration_label:SetFontStyle(self.duration_font_outline)
    self.duration_label:SetMultiline(false)
    self.duration_label:SetMarkupEnabled(true)

	self:Format()
end

function Timer:UpdateBar()
	if self.max_duration > 0 then
		local progress = math.clamp(self.current_duration / self.max_duration, 0, 1)
		if self.increasing then
			progress = 1 - progress
		end

		if self.show_duration then
			if self.increasing then
				self.duration_label:SetText(self:StringifyDuration(self.max_duration - self.current_duration))
			else
				self.duration_label:SetText(self:StringifyDuration(self.current_duration))
			end
		end
		
		-- Update the bar postion and size
		self.bar:SetPosition(self.bar_left_to_right and self.bar_background:GetWidth() - progress * self.bar_background:GetWidth() or 0 , 0)
		self.bar:SetSize(progress * self.bar_background:GetWidth(), self.bar_background:GetHeight())

		-- Update the warning colour
		if self.useWarning and self.current_duration < self.warning_threshold then
			local timerLeft = (self.warning_threshold - self.current_duration) / self.warning_threshold
			self.bar:SetBackColor(Turbine.UI.Color(
				self.bar_background_colour.A,
				math.abs(self.warning_colour.R - self.bar_background_colour.R) * timerLeft, 
				math.abs(self.warning_colour.G - self.bar_background_colour.G) * timerLeft, 
				math.abs(self.warning_colour.B - self.bar_background_colour.B) * timerLeft))
		end
	end
end

function Timer:GetRemainingTime() 
	if self.increasing then
		return self.max_duration - self.current_duration 
	else
		return self.current_duration 
	end
end

function Timer:GetDuration() return self.max_duration end

function Timer:StartTimer()
	if not self.update_callback then
		self.update_callback = AddCallback(Updater, "ExecuteUpdate", self.UpdateFunction)
		self.UpdateEvent(self, self.TimerStarted, {["State"] = "started"})
	end
end

function Timer:StopTimer()
	if self.update_callback then
		RemoveCallback(Updater, "ExecuteUpdate", self.update_callback)
		self.update_callback = nil
	end
end

function Timer:UpdateFunction(sender, delta)
	self.current_duration = self.current_duration - delta
	self:UpdateBar()

	if self.current_duration <= 0 then
		self:StopTimer()
		self.UpdateEvent(self, self.TimerEnded, {["State"] = "stopped"})
	end
end

function Timer:Trigger(args)
	if args.Start_Timer then
		self:SetTimer(args.Duration, args.Max_Duration)
		self:StartTimer()
	end

	if args.Pause then
		self:StopTimer()
	end

	if args.Start then
		self:StartTimer()
	end
end

function Timer:SetTimer(duration, maxDuration)
	self.current_duration = duration
	self.max_duration = maxDuration
end

function Timer:Unload()
	self:StopTimer()
	self:SetVisible(false)
    local controlList = self:GetControls()
    for i = 1, controlList:GetCount() do
        controlList:RemoveAt(1)
    end
    self:SetParent(nil)
end

function Timer:Format()
	-- Main UI
	self:SetSize(self.width, self.height)
	self:SetOpacity(self.overal_opacity)

	-- Border
    self.border_window:SetPosition(0, 0)
    self.border_window:SetSize(self:GetWidth(), self:GetHeight())

	-- Inner Timer
	self.inner_window:SetPosition(self.border_size, self.border_size)
    self.inner_window:SetSize(self.border_window:GetWidth() - 2 * self.border_size, self.border_window:GetHeight() - 2 * self.border_size)

	-- Bar
	self.bar_border_window:SetPosition(self.left_icon and self.inner_window:GetHeight() or 0, self.inner_window:GetHeight() - self.bar_height_percentage * self.inner_window:GetHeight())
	self.bar_border_window:SetSize(self.inner_window:GetWidth() - (self.left_icon and self.inner_window:GetHeight() or 0) - (self.right_icon and self.inner_window:GetHeight() or 0), self.bar_height_percentage * self.inner_window:GetHeight())

	self.bar_background:SetPosition(self.bar_border_size, self.bar_border_size)
	self.bar_background:SetSize(self.bar_border_window:GetWidth() - 2 * self.bar_border_size, self.bar_border_window:GetHeight() - 2 * self.bar_border_size)

	self.bar:SetPosition(0, 0)
	self.bar:SetSize(0, self.bar_background:GetHeight())

	-- Left Icon
	if self.left_icon then
		self.left_icon_control:SetOpacity(self.icon_opacity)
		self.left_icon_control:SetStretchMode(2)
		self.left_icon_control:SetStretchMode(1)
		self.left_icon_control:SetSize(self.inner_window:GetHeight(), self.inner_window:GetHeight())
	end

	-- Right Icon
	if self.right_icon then
		self.right_icon_control:SetOpacity(self.icon_opacity)
		self.right_icon_control:SetStretchMode(2)
		self.right_icon_control:SetStretchMode(1)
		self.right_icon_control:SetPosition(self.inner_window:GetWidth() - self.inner_window:GetHeight(), 0)
		self.right_icon_control:SetSize(self.inner_window:GetHeight(), self.inner_window:GetHeight())
	end

	-- Timer Text
	self.text_window:SetPosition(self.left_icon and self.inner_window:GetHeight() or 0, 0)
    self.text_window:SetSize(self.inner_window:GetWidth() - (self.left_icon and self.inner_window:GetHeight() or 0) - (self.right_icon and self.inner_window:GetHeight() or 0), self.inner_window:GetHeight())
    self.text_window:SetOpacity(self.text_opacity)

	self.text_label:SetPosition(0, 0)
    self.text_label:SetSize(self.text_window:GetWidth(), self.text_window:GetHeight())

	-- Duration Text
	self.duration_window:SetPosition(self.left_icon and self.inner_window:GetHeight() or 0, 0)
    self.duration_window:SetSize(self.inner_window:GetWidth() - (self.left_icon and self.inner_window:GetHeight() or 0) - (self.right_icon and self.inner_window:GetHeight() or 0), self.inner_window:GetHeight())
    self.duration_window:SetOpacity(self.duration_opacity)

	self.duration_label:SetPosition(0, 0)
    self.duration_label:SetSize(self.text_window:GetWidth(), self.text_window:GetHeight())
end

function Timer:StringifyDuration(duration, round)
    if not duration then return "" end
    if round then
        if duration <= 60 then
            return string.format("%." .. round .. "f", duration)
        elseif duration <= 3600 then
            return string.format("%i:%02." .. round .. "f", duration / 60, duration % 60, duration % 60)
        else
            return string.format("%i:%02d:%02." .. round .. "f", duration / 3600, (duration % 3600) / 60, duration % 60)
        end
    else
        if duration <= 60 then
            return string.format("%i", duration)
        elseif duration <= 3600 then
            return string.format("%i:%02d", duration / 60, duration % 60)
        else
            return string.format("%i:%02d:%02d", duration / 3600, (duration % 3600) / 60, duration % 60)
        end
    end
end